#! /usr/bin/env python

import argparse
import os.path, time
from modules.utilities import emit_list, get_processes, do_command 
from modules.utilities import do_parse_2D, do_command_stdout, simulate 
from modules.utilities import print_traces, dict_param, effective_param

debug = False
verbose = False

dict_Default = {'name': 'basicanalysis', 
                          'ppn': '1', 'sc': 'weak', 
                          'bw': '1024', 'lat': '4', 
                          'freq': '2.9', 'sim': 'time'}

# Points to the automatic_scripts folder 
SCRIPTS_DIR= os.environ['BASIC_ANALYSIS_HOME']
TMP_DIR = './tmp'
 
### HELPER FUNCTIONS ###
def get_parameters():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", metavar='indat_modelfactors.ban', nargs='?', help = "Configuration file")
    parser.add_argument('-sim', choices=['time', 'cycles'], 
                        help=("Simulations based on time (default) or cycles"), 
                        default=None, nargs='?')
    parser.add_argument('-ppn',  type=int, nargs='?', 
                        help=("Processes per node. Dimemas Simulations"),
                        default=None)
    parser.add_argument('-lat',  type=int, nargs='?', 
                        help=("Nominal Latency for simulations, required with 'cycles'"),
                        default=None)
    parser.add_argument('-bw', type=int, nargs='?',
                        help="Nominal BW for simulations, required with 'cycles'",
                        default=None)
    parser.add_argument('-fq', type=float, nargs='?',
                       help="Processor's Speed in GHz, i.e. 2.9, required with 'cycles'",
                       default=None)
    parser.add_argument('-sc', choices = ['weak', 'strong'], nargs='?', 
                        help = 'Weak or Strong Scaling Analysis',
                        default=None)
    parser.add_argument('-phase', nargs='?', 
                        help="Name of the phase represented in traces",
                        default=None)
    parser.add_argument('-t', nargs='+', required = True, 
                        help="List with the names of the traces")

    args = parser.parse_args()
    
    if (args.i != None):
        confFile = True
        indat = open(args.i, 'r')
        dargs = dict_param(indat)
        
    else:
        print "Running without configuration file"
        print "--- using default values or inputs from command line"
        confFile = False
        dargs = {}
        dargs ['bw'] = None
        dargs['lat'] = None
        dargs['sc'] = None
        dargs['freq'] = None
        dargs['ppn'] = None
        dargs['sim'] = None
        dargs['name'] = None
        if (debug):
            print dargs
    
    traces_list = args.t
 
    d_com_line = {}
    
    d_com_line['bw'] = args.bw
    d_com_line['lat'] = args.lat
    d_com_line['sc'] = args.sc
    d_com_line['freq'] = args.fq
    d_com_line['ppn'] = args.ppn
    d_com_line['sim'] = args.sim
    d_com_line['name'] = args.phase

    name_params = ['bw', 'sim', 'lat', 'sc', 'ppn', 'name', 'freq']
    final_params = effective_param(dargs, d_com_line, name_params, confFile, dict_Default)
    bw = final_params['bw'] 
    latency = final_params['lat']
    scaling = final_params['sc']
    cpufreq =final_params['freq']
    ppn = final_params['ppn']
    dim_sim = final_params['sim']
    phase_name = final_params['name']
            
    return bw, latency, scaling, cpufreq, ppn, dim_sim, traces_list, phase_name

def initialize_values():
    paraver_files = []
    perproc = []
    analyzed = []
    IPC = []
    LB_cst = 1.0
    Tcpu_cst = 0.0
    elems = 0
    accum = 0.0
    Tcpu_it = 0
    T0 = 0.0
    min_procs = 1E12
    
    return (paraver_files, perproc, analyzed, IPC, LB_cst, Tcpu_cst, elems, 
            accum, Tcpu_it, T0, min_procs)
    
def meas_path(dim_sim):
    if (dim_sim == 'time'):
        measured_file    = './'    + phase_name + '.T.csv'
        fun_facts_file = './fundamental_factors.' + phase_name + '.T.csv'
    elif (dim_sim == 'cycles'):
        measured_file    = './'    + phase_name + '.C.csv'
        fun_facts_file = './fundamental_factors.' + phase_name + '.C.csv'
    else:
        print "Not valid option!!!!"
    
    print "File %s stores resulting model factors" %measured_file 
    
    return (measured_file, fun_facts_file)

def get_processor(traces_list, min_procs, paraver_files):
    
    for name in sorted(traces_list):
        processes = get_processes(name)
        paraver_files.append([processes, name])
        if (debug):
            print paraver_files
        if (processes < min_procs):
            min_procs = processes
    
    return (min_procs)

def get_trace_data(trace_data):
    
    #processes = trace_data[0]
    processes = get_processes(trace_data[1])
    trace = trace_data[1]

    occur = trace.count('/')
        
    if (verbose):
        print trace
        print processes
    
    trace_path = '' 
    if (occur >= 1):
        paths = trace.split('/')
        fields_path = int(len(paths))
        if(debug):
            print fields_path
        trace_name = paths[len(paths)-1]
        
        for i in range(len(paths)-1):
            if(debug):
                print paths[i]
            trace_path += paths[i] + '/'
    elif (occur == 0):
        trace_name = trace
    else:
        print "ERROR!!!!!"
        
    trace_path += trace_name
    if(debug):
        print trace_path
    
    print "\n\n"
    print "_____________________________________________________________"
    print "============================================================="
    print 'Analyzing: \n ', trace_name
    print "_____________________________________________________________"
    print "============================================================="

    return (processes, trace_name, trace_path)

def get_useful_time(trace_path):
    duration  = 0.0
    with open (trace_path) as f:
        for line in f:
            words = line.split(':')
            if words[0] == str(2):
                event_time = float(words[5])
                if event_time > duration:
                    duration = event_time
                if words[0] == str(1):
                    if words[7] == str(1):
                        event_time = float(words[6])
                        if event_time > duration:
                            duration = event_time 
        
        duration = duration/1.0e8
        
        return duration
    
def gen_prv2dim(dim_trace, trace_path, dim_sim):
    Dimemas_file = TMP_DIR + '/' + dim_trace[6:-4] + '.TRANS_Dimemas.out'
    
    fd=open(Dimemas_file, 'w')
    
    if (dim_sim == 'time'):   
        print "Dimemas trace with time"
        do_command_stdout(['prv2dim', '-n', '-s', trace_path, dim_trace], fd, verbose)
    elif (dim_sim == 'cycles'):
        print "Dimemas trace with cycles"
        cpu_ratio=float(1e-9/cpufreq)
            
        if (debug):
            print cpu_ratio
                
        filter_event = '42000059'
        filter_cyc = str(cpu_ratio)
        line_com = filter_event + ',' + filter_cyc
        do_command_stdout(['prv2dim', '-n', '-b', line_com, trace_path, 
                           dim_trace], fd, verbose)
    else:
        print "ERROR using prv2dim!!!!!"
    
    return dim_trace

def setup_prv2dim(dim_sim, trace_name, trace_path):
    
    if (dim_sim == 'time'):
        dim_trace = TMP_DIR + '/' + trace_name[:-4]+'.dim'
        dim_trace_path = trace_path[:-4] + '.dim' 
        print trace_path
    elif (dim_sim == 'cycles'):
        dim_trace = TMP_DIR + '/' + trace_name[:-4] + '_cyc.dim'
        dim_trace_path = trace_path[:-4] + '_cyc.dim' 
    else:
        print "PRV2DIM --- Option not valid!!!"

    mod_trace = time.ctime(os.path.getctime(trace_path))
    if (verbose):
        print "last time changed original_trace %s"%mod_trace     
    
    if (os.path.exists(dim_trace) or os.path.exists(dim_trace_path)):
        if (os.path.exists(dim_trace_path)):
            do_command(['cp', dim_trace_path, TMP_DIR], verbose)
            
        #print dim_trace
        mod_sim_trace = time.ctime(os.path.getctime(dim_trace))
        if (verbose):
            print "Last time changed sim_trace %s"%mod_sim_trace
        #TODO: there's something wrong with the time of .dim, because always gets
        #into the else, and regenerates a dim file
        #if (mod_sim_trace < mod_trace):
        #    print "[!] Dimemas trace already exists. Reusing it"
        #else:
        #    print "[!] Dimemas trace is older than *.prv. Generating a new one"
        #    dim_trace = gen_prv2dim(dim_trace, trace_path, dim_sim)
    else:
        print"Generating Dimemas trace (prv2dim)"
        dim_trace = gen_prv2dim(dim_trace, trace_path, dim_sim)
                    
    return dim_trace

def sim_ideal(dim_sim, trace_name, trace_path, Next_trace):
    Next_trace = False
    
    if (dim_sim == 'time'):
        D_trace_name = TMP_DIR + '/' + '.'.join(['D', trace_name[:-4], 
                                                 'ideal.prv' ])
    elif (dim_sim == 'cycles'):
        D_trace_name = TMP_DIR + '/' + '.'.join(['D', trace_name[:-4], 
                                                 'cyc_ideal.prv' ])
    else:
        print "NAME of Ideal -- Option not valid"
    
    CPUr = 1.0
    in_links = 10
    out_links = 10

    Tideal = (simulate(dim_trace, 0.000000, 0.000000, CPUr, in_links, out_links, 
                      D_trace_name, TMP_DIR, ppn, verbose))
    
    command = ['paramedir', '-npr', D_trace_name ]
    command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_mpi_stats.cfg',
                             TMP_DIR + '/2dp_mpi_stats.D.stats'])
    command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_useful.cfg',
                             TMP_DIR + '/2dp_useful.D.stats'])
    do_command( command, verbose)

    #Generates a simulated trace with nominal values for latency and BW
    if (dim_sim == 'cycles'):
        D_trace_name = TMP_DIR + '/' + '.'.join(['D',trace_name[:-4],'cyc_based',
                                                 ''.join(['BW', str(bw)]),''.join(['L',str(latency)]),'prv'])
        exec_time_0_n = simulate (dim_trace, float(bw), float(latency)/1000000, CPUr,
                                                 in_links, out_links, D_trace_name, TMP_DIR, ppn,
                                                 verbose)
        trace_path = D_trace_name

        if (debug):
            print Tideal, exec_time_0_n, trace_path
  
    if Tideal == (-1.0):
        Next_trace = True

    return Next_trace
    

### MAIN PROGRAM ###
if __name__ == "__main__":
    
    # Get parameters 
    (bw , latency, scaling, cpufreq, ppn, 
     dim_sim, traces_list, phase_name) = get_parameters()

    # Initialize variables 
    (paraver_files, perproc, analyzed, IPC, LB_cst, Tcpu_cst, elems, 
            accum, Tcpu_it, T0, min_procs) = initialize_values()
    
    # Set path and file name to measured file
    measured_file, fun_facts_file = meas_path(dim_sim)

    # To know which traces will be evaluated
    if (verbose):
        print "Traces to analyze: %s " %traces_list

    # Get number of processes in the trace 
    min_procs = get_processor(traces_list, min_procs, paraver_files)
    
    if (verbose):
        print paraver_files

    # Creates temporary file 
    if os.path.exists(TMP_DIR):
        print "There's a TMP Directory available. Reusing it!"
    #   os.system("rm -rf ./tmp")
    else:
        do_command(['mkdir', './tmp'], verbose)
    
    ### BEGINS MAIN LOOP ###
    total_traces = len(paraver_files)
    list_traces = (sorted(paraver_files))
    processed = []
    excluded = []
    met = []
    fun_facts = []
    
    proc_count = 0
    exc_count = 0
    
    for trace_data in list_traces:    
    #  Gets trace name 
    #"Evaluates if there is a path in the name of the trace"
        (processes, trace_name, trace_path) = get_trace_data(trace_data)

        # Useful Time
        duration = get_useful_time(trace_path)
        
        #Prepare everything for the prv2dim 
        dim_trace = setup_prv2dim(dim_sim, trace_name, trace_path)
    
        # Paramedir with a trace with unlimited BW and No-Latency
        nt = False
        nt = sim_ideal(dim_sim, trace_name, trace_path, nt)
        
        if (nt == False):
            processed.append(trace_name)
            proc_count += 1
        #================================================
        #  Paramedir with traces
        #================================================
            command = ['paramedir', '-npr', trace_path ]
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_mpi_stats.cfg', 
                             TMP_DIR+'/2dp_mpi_stats.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_useful.cfg', 
                             TMP_DIR+'/2dp_useful.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_mpi_time_stats.cfg', 
                             TMP_DIR+'/2dp_mpi_time_stats.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_total_mpi_time_stats.cfg', 
                             TMP_DIR+'/2dp_total_mpi_time_stats.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_ui.cfg', 
                             TMP_DIR+'/2dp_ui.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_ipc.cfg', 
                             TMP_DIR+'/2dp_ipc.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_mips.cfg', 
                             TMP_DIR+'/2dp_mips.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_cyclesperus.cfg', 
                             TMP_DIR+'/2dp_cyclesperus.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2d_cycles_useful.cfg', 
                             TMP_DIR+'/2d_cycles_useful.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_maxbytes_Mpicalltype.cfg', 
                             TMP_DIR+'/2dp_maxbytes_Mpicalltype.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_num_Mpicalltype.cfg', 
                             TMP_DIR+'/2dp_num_Mpicalltype.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_bytes_sent.cfg', 
                             TMP_DIR+'/2dp_bytes_sent.stats'])
            command.extend ([ SCRIPTS_DIR+'/cfgs/2dp_bytes_Mpicalltype.cfg', 
                             TMP_DIR+'/2dp_bytes_Mpicalltype.stats'])
            do_command( command, verbose)
            
            max_time = float(do_parse_2D('./tmp/2dp_mpi_time_stats.stats', 
                                           'Maximum', 1))/1000000
            usefulTime = float(do_parse_2D('./tmp/2dp_mpi_time_stats.stats', 
                                           'Average', 1))/1000000
            MPItime = float(do_parse_2D('./tmp/2dp_total_mpi_time_stats.stats', 
                                        'Average', 2))/1000000
            efficiency = float(do_parse_2D('./tmp/2dp_mpi_stats.stats', 
                                           'Average', 1))/100
            LB = float(do_parse_2D('./tmp/2dp_mpi_stats.stats', 
                                   'Avg/Max', 1))
            accum += usefulTime
            uLB = float(do_parse_2D('./tmp/2dp_mpi_stats.D.stats', 
                                    'Maximum', 1))/100
            CommEff = float(do_parse_2D('./tmp/2dp_mpi_stats.stats', 
                                        'Maximum', 1))/100
#             uLB = float(do_parse_2D('./tmp/2dp_useful.D.stats', 
#                                     'Maximum', 2))/100
#             CommEff = float(do_parse_2D('./tmp/2dp_useful.stats', 
#                                         'Maximum', 2))/100
            Transfer = CommEff/uLB
            err = (efficiency - LB*uLB*Transfer)/efficiency
            ui = float(do_parse_2D('./tmp/2dp_ui.stats', 'Average', 1))
            Total_ui = float(do_parse_2D('./tmp/2dp_ui.stats', 'Total', 1))
            print "Total Instructions", Total_ui
            LB_ui = float(do_parse_2D ('./tmp/2dp_ui.stats', 'Avg/Max', 1))
            ipc = float(do_parse_2D('./tmp/2dp_ipc.stats', 'Average', 1))
            LB_ipc = float(do_parse_2D ('./tmp/2dp_ipc.stats', 'Avg/Max', 1))
            mips = float(do_parse_2D('./tmp/2dp_mips.stats', 'Average', 1))
            LB_mips = float(do_parse_2D ('./tmp/2dp_mips.stats', 'Avg/Max', 1))
            cyclesperus = float(do_parse_2D('./tmp/2dp_cyclesperus.stats', 
                                            'Average', 1))
            cycles_useful = float(do_parse_2D('./tmp/2d_cycles_useful.stats', 
                                              'Average', 1))
            p2p_calls = float(do_parse_2D('./tmp/2dp_num_Mpicalltype.stats', 
                                          'Maximum', 1))
            #p2p_calls = float(do_parse_2D('./tmp/2dp_num_Mpicalltype.stats', 
            #                              'Average', 1))
            LB_p2p_calls = float(do_parse_2D('./tmp/2dp_num_Mpicalltype.stats', 
                                             'Avg/Max', 1))
#             p2p_bytes  = float(do_parse_2D('./tmp/2dp_bytes_sent.stats', 
#                                         'Average', 1))
            p2p_bytes  = float(do_parse_2D('./tmp/2dp_bytes_sent.stats', 
                                           'Maximum', 1))
            LB_p2p_bytes = float(do_parse_2D('./tmp/2dp_bytes_sent.stats', 
                                             'Avg/Max', 1))
            col_calls = float(do_parse_2D('./tmp/2dp_num_Mpicalltype.stats', 
                                          'Average', 2))
            col_bytes = float(do_parse_2D('./tmp/2dp_bytes_Mpicalltype.stats', 
                                          'Average', 2))
            max_col_bytes = float(do_parse_2D('./tmp/2dp_maxbytes_Mpicalltype.stats', 
                                              'Average', 2))    
            LB_col_bytes = float(do_parse_2D('./tmp/2dp_bytes_Mpicalltype.stats', 
                                             'Avg/Max', 2))
        
            predicted = float((usefulTime)/efficiency)
            
            par_Eff = float(LB * uLB * Transfer)
            IPC.append(float(ipc))
            
            elems += 1
            if elems == 1:
                T0 = predicted
                time0 = duration
                inst0 = ui
                ipc0 = ipc
                mips0 = mips
                cyc0  = cyclesperus
                cycus0  = cycles_useful
                tui0 = Total_ui
                if (p2p_bytes != 0):
                    p2p_bytes_0 = p2p_bytes
                else:
                    p2p_bytes_0 = 1
                if (p2p_calls != 0):
                    p2p_calls_0 = p2p_calls
                else:
                    p2p_calls_0 = 1
                if (col_bytes != 0):    
                    col_bytes_0 = col_bytes
                else:
                    col_bytes_0 = 1
                if (col_calls != 0):
                    col_calls_0 = col_calls
                else:
                    col_calls_0 = 1
                if (max_col_bytes != 0):
                    max_col_bytes_0 = max_col_bytes
                else:
                    max_col_bytes_0 = 1
            
            predScal = float (T0 / predicted)
        
            #p2p ratios
            if (p2p_calls != 0):
                eff_p2p_calls = float(p2p_calls/p2p_calls_0)
            else:
                eff_p2p_calls = 0
            if (p2p_bytes != 0):
                eff_p2p_bytes = float(p2p_bytes/p2p_bytes_0)
            else:
                eff_p2p_bytes = 0
            
            #collective ratios
            if (col_calls != 0):
                eff_col_calls = float(col_calls_0/col_calls)
            else:
                eff_col_calls = 0
            if (col_bytes != 0):
                eff_col_bytes = float(col_bytes_0/col_bytes)
            else:
                eff_col_bytes = 0
            if (max_col_bytes != 0):
                eff_max_col_bytes = float(max_col_bytes_0/max_col_bytes)
            else:
                eff_max_col_bytes = 0
            
            if (scaling == 'weak'):
                if (Total_ui != 0):
                    eff_inst = (tui0 * (processes/min_procs))/Total_ui
                else:
                    eff_inst = 0.0
                #efficiency time
                speedup = float (time0 / duration)
            elif (scaling == 'strong'):
                if (Total_ui != 0):
                    eff_inst = float(tui0/Total_ui)
                else:
                    eff_inst = 0.0
                speedup = float(time0*min_procs)/(processes*duration)
            
            eff_ipc = 0.00
            eff_mips = 0.00
            eff_cyc = 0.00
            eff_cyc_use = 0.00
            
            Global_eff = 0.00
        
        #================================================
        #	Printing fundamental factors per trace  
        #================================================
            print "---------------------------------------------------------"
            print "    Fundamental factors for trace with %d processes"%processes
            print "---------------------------------------------------------\n"
            print "LB         " + " " + "Trf       " + " " + "Ser       " + " " + "ParEff"
            print "%.10f"%LB , " " , "%.10f"%Transfer , " " , "%.10f"%uLB , " " , "%.10f"%efficiency
    
            print "---------------------------------------------------------"
            print "Load Balance   : %.10f"%LB
            print "Serialization  : %.10f"%uLB
            print "Transfer       : %.10f"%Transfer
            print "Par. Efficiency: %.10f"%efficiency
            print "---------------------------------------------------------"
        #================================================
        #  Final store into arrays, dictionary and .csv file
        #================================================
            
            Pro_3 = float(int(processes)**(1.0/3.0))
            met = [int(processes), float(format(efficiency, '.10f')), float(format(LB,'.10f')),
                    float(format(uLB, '.10f')), float(format(Transfer,'.10f')), 
                    float(format(par_Eff,'.10f')), float(format(CommEff, '.10f')),
                    abs(float(format(err, '.10f'))), float(format(max_time,'.10f')),
                    float(format(duration,'.10f')), float(format(ui/1000000, '.10f')), 
                    float(format(ipc,'.6f')), float(format(eff_ipc, '.10f')), float(format(eff_inst,'.10f')), 
                    float(format(speedup, '.10f')), float(format(eff_mips,'.10f')), 
                    float(format(mips,'.10f')), float(format(cycles_useful, '.10f')), 
                    float(format(eff_cyc_use,'.10f')), float(format(cyclesperus, '.10f')), 
                    float(format(eff_cyc, '.10f')), float(format(predicted,'.10f')), 
                    float(format(predScal,'.10f')), float(format(LB_ui,'.10f')), 
                    float(format(LB_ipc,'.10f')), float(format(Pro_3,'.10f')), float(1.000000), 
                    float(format(p2p_calls, '.10f')), float(format(eff_p2p_calls, '.10f')), 
                    float(format(LB_p2p_calls,'.10f')), float(format(p2p_bytes, '.10f')), 
                    float(format(eff_p2p_bytes, '.10f')), float(format(LB_p2p_bytes,'.10f')), 
                    float(format(col_calls, '.10f')), float(format(eff_col_calls, '.10f')), 
                    float(format(col_bytes, '.10f')), float(format(eff_col_bytes, '.10f')), 
                    float(format(max_col_bytes, '.10f')), float(format(eff_max_col_bytes, '.10f')),
                    float(format(LB_col_bytes, '.10f')), float(format(Global_eff, '.10f'))]
            analyzed.append(met)
            
            met = [int (processes), float(duration), 
                   float(LB), float(uLB), float(Transfer),
                   float(par_Eff)]
            fun_facts.append(met)
                    
        else:
            exc_count += 1
            print "There are warnings %d"%exc_count
            excluded.append(trace_name)
        
        names_csv = "proc, TCPU, LB, Ser, Trf, ParEff\n"
        
        emit_list(fun_facts_file, names_csv, fun_facts, verbose)
        #---------------------------------------------------------------------------------------------------------
        #END LOOP
        #---------------------------------------------------------------------------------------------------------
    #================================================
    #    Saves model factors into measured_<name_of_the_phase>.csv file
    #================================================
    if (len(processed) > 0):
        IPC = []
        MIPS = []
        CYCUS = []
        CYCS =[]
        
        for processes in sorted(analyzed):
            values = processes
            IPC.append(values[11])
            MIPS.append(values[16])
            CYCUS.append(values[17])
            CYCS.append(values[19])
        
            if (debug):
                print 'IPC from array', IPC
                print 'MIPS from array', MIPS
                print 'CYCUS from array', CYCUS
                print 'CYCS from array', CYCS
        
        ipc0 = (IPC[0])
        mips0 = (MIPS[0])
        cycus0 = (CYCUS[0])
        cyc0 = (CYCS[0])
        
        for processes in sorted(analyzed):
            values = list(processes)
            ipc = values[11]
            mips = values[16]
            cycles_useful = values[17]
            cycles = values[19]
        
            if (ipc0 != 0):
                eff_ipc = float (ipc / ipc0)
            else:
                eff_ipc = 0.0
                
            if (mips0 != 0):
                eff_mips = float (mips / mips0)
            else:
                eff_mips = 0.0
                
            if (cyc0 != 0):
                eff_cyc = float(cycles/cyc0)
            else:
                eff_cyc = 0.0
                
            if (cycus0 != 0):
                eff_cyc_use = float(cycles_useful/cycus0)
            else:
                eff_cyc_use = 0.0
             
            processes[12] = eff_ipc
            processes[15] = eff_mips
            processes[18] = eff_cyc_use
            processes[20] = eff_cyc
        
        for processes in sorted(analyzed):
            values = list(processes)
            eff_ipc = values[12]
            eff_inst = values[13]
            par_eff = values[5]
            
            Global_eff = eff_ipc * eff_inst * par_eff
             
            processes[40] = Global_eff
        
        sorted(analyzed)
        
        names_csv = "proc, Eff, LB, Ser, Trf, ParEff, CommEff, err, TCPU, "
        names_csv = names_csv + "Time, ui, ipc, eff_ipc, eff_inst, eff_time, eff_mips, "
        names_csv = names_csv + "mips, cycles_useful, eff_cyc_useful, cyclesperus, "
        names_csv = names_csv + "eff_cyc, prediction, Scaled_eff, LB_ui, LB_ipc, " 
        names_csv = names_csv + "proc_13, Sche, p2p_calls, eff_p2p_calls, "
        names_csv = names_csv + "LB_p2p_calls, p2p_bytes,  eff_p2p_bytes, "
        names_csv = names_csv + "LB_p2p_bytes, col_calls, eff_col_calls, col_bytes, "
        names_csv = names_csv + "eff_col_bytes, max_col_bytes, eff_max_col_bytes, "
        names_csv = names_csv + "LB_col_bytes, Global_eff\n"
        
        emit_list(measured_file, names_csv, analyzed, verbose)
        
        #================================================
        #    Builds GNUPlot file to be shown in Paraver
        #================================================
        
        from modules.genGNUPlot import genGNUPlotfile, genSpeedupGraph, genEffGraph
        from modules.genGNUPlot import genLBGraph, genRateGraph, genTimeGraph
        from modules.genGNUPlot import genCyclesGraph, genCollectivesGraph, genP2PGraph
        genGNUPlotfile(measured_file, dim_sim)
        genSpeedupGraph(measured_file)
        genEffGraph(measured_file)
        genLBGraph(measured_file)
        genRateGraph(measured_file)
        genTimeGraph(measured_file)
        genCyclesGraph(measured_file)
        genCollectivesGraph(measured_file)
        genP2PGraph(measured_file)
        
    print "\n\n\nTraces evaluated: ", proc_count
    if (len(processed) > 0):
        it = iter(sorted(processed))
        print_traces(it)
    else:
        print "There are not traces processed."
    
    print "Traces excluded: ", exc_count
    if (len(excluded) > 0):
        it = iter(sorted(excluded))
        print_traces(it)
    else:
        print "There are not traces excluded."  

    #================================================
    #    Cleaning Up
    #================================================
    #os.system("rm -r ./tmp")
    #---------------------------------------------------------------------------------------------------------
    #END model_factors.py  
    #---------------------------------------------------------------------------------------------------------
