#! /usr/bin/env python
######################################################################
# method to perform the projection of measured metrics
# Calls the corresponding fits
# returns a list
######################################################################
from lmfit import minimize, Parameters, report_errors, report_fit
from scipy.optimize import curve_fit
import numpy as np

debug = False                    #Flag to activate prints to debug

#=====================================================================
def get_initial_values(evolP, fitFactor, P, measFactor, funFactor):

    Pconv = convert_List (evolP, P)
    if (fitFactor == 'pipeline'):
        if (funFactor == 'Load Balance'):
            print "Fitting not valid for %s"%funFactor
            elem_0, f_elem = [0.0, 0.0]
        else:
            if (debug):
                print "Pipeline for %s........."%funFactor
            elem_0, f_elem = minimize_pipe(np.array(measFactor), Pconv)

    elif (fitFactor == 'amdahl'):
        if (debug):
            print "Amdahl for %s..........."%funFactor
        elem_0, f_elem = minimizing(np.array(measFactor), Pconv)

    elif (fitFactor == 'log'):
        if (debug):
            print "Logaritmic for %s............."%funFactor
        popt, pcov = curve_fit(log_fit, Pconv, np.array(measFactor))
        elem_0, f_elem = Min_log(np.array(measFactor), Pconv)
    else:
        print "Fitting not valid for %s"%funFactor
        elem_0, f_elem = [0.0, 0.0]
        
    return elem_0, f_elem

#=====================================================================
def get_initial_values_single(fitFactor, measFactor, funFactor):
    if (fitFactor == 'min'):
        if (debug):
            print "Minimum for %s............"%funFactor
        constFactor = Min_fit(measFactor)
    elif (fitFactor == 'avg'):
        if (debug):
            print "Average for %s............"%funFactor
        constFactor = Avg_fit(measFactor)
    
    return constFactor

#=====================================================================
def projection (Evol_P, fitFactor, procs, resFactor, funFactor):
    
    P = conv_P(Evol_P, procs)
    
    if (fitFactor == 'pipeline'):
        if (debug):
            print "Pipeline for %s"%funFactor
        resultFactor = f_pipe(P, resFactor)
    if (fitFactor == 'amdahl'):
        if (debug):
            print "Amdahl for %s"%funFactor
        resultFactor   = f_amdahl(P, resFactor)
    if (fitFactor == 'log'):
        if (debug):
            print "Log for %s"%funFactor
        #NOT TESTED ENOUGH
        resultFactor = log_fit2(P, resFactor)
    if ((fitFactor != 'pipeline') and (fitFactor != 'amdahl') 
        and (fitFactor != 'log')):
        print "Fitting option not valid"
        resultFactor = 0.0
    return resultFactor

#=====================================================================

#TODO: get this function to use only arrays
def conv_P (Evol_P, procs):
    P = 0
    if (Evol_P == 'cubic'):
        #Cubic Root of P
        P     = np.power(procs, (1.0/3.0))
    elif (Evol_P == 'squared'):
        #Square root of P
        P =  np.power(procs,(1.0/2.0))
    elif (Evol_P == 'log'):
        # Lob base 2 of P
        P = np.log2(procs)
    else:
        #Linear
        P     = procs
    return P

#=====================================================================
def convert_List (Evol_P, x):
    nax = []
    if (Evol_P == 'cubic'):
        #Cubic Root of P
        nax   = np.power(x,(1.0/3.0))
    elif (Evol_P == 'squared'):
        #Square root of P
        nax =  np.power(x,(1.0/2.0))
    elif (Evol_P == 'log'):
        # Lob base 2 of P
        nax = np.log2(x)
    else:
        #Linear
        nax = x

    return nax
#=====================================================================

 
#=====================================================================
#  Pipeline-based fitting 
#=====================================================================
def f_pipe(procs, var_elem):
    e_elem = var_elem[1]
    f_elem = var_elem[0]
    return (e_elem * procs)/((1-f_elem) + (f_elem*(2*procs-1)))

def res_pipe(params, procs, measured):
    e_elem = params['e_elem'].value
    f_elem = params['f_elem'].value
    var_elem = [f_elem, e_elem]
    model = f_pipe(procs, var_elem)
    if (debug):
        print sum((measured -model)*(measured - model))
    return (measured - model)

def minimize_pipe(meas, x):
    params = Parameters()
    params.add('e_elem', value= 1.00000000, min=0.000000, max=1.00000000)
    params.add('f_elem', value= 1.00000000, min=0.000000, max=1.00000000)
    if(debug):
        test = res_pipe(params, x, meas)
        print params, x, meas, test
    out = minimize(res_pipe, params, args=(x,meas))
    out.leastsq()
    e_elem = out.params['e_elem'].value
    f_elem = out.params['f_elem'].value
    if(debug):
        report_errors(params)
        report_fit(out.params)
    return e_elem, f_elem
  
#=====================================================================

#=====================================================================
#  Amdahls-based fitting 
#=====================================================================
def f_amdahl(procs, var_elem):
    e_elem = var_elem[1]
    f_elem = var_elem[0]
    aux = e_elem/(f_elem + (1-f_elem)*procs)
    return aux
#    return e_elem/(f_elem + (1-f_elem)*procs)

# def f_amdahl_inst(procs, var_elem):
#     aux = 0.0000000
#     e_elem = var_elem[1]
#     f_elem = var_elem[0]
#     print "PROCS: ", procs
#     print "e_elem: ",e_elem
#     print "f_elem: ",f_elem
#     aux = e_elem/(f_elem + (1-f_elem)*procs)
#     print "effic: ",aux
#     return aux
#    return e_elem/(f_elem + (1-f_elem)*procs)

def residual(params, procs, measured):
    e_elem = params['e_elem'].value
    f_elem = params['f_elem'].value
    var_elem = [f_elem, e_elem]
    model = f_amdahl(procs, var_elem)
    if (debug):
        print sum((measured-model)*(measured-model))
    return (measured - model) 

def minimizing(meas, x):
    params = Parameters()
    params.add('e_elem', value= 1.00000000, min=0.00000000, max=1.00000000)
    params.add('f_elem', value= 1.00000000, min=0.00000000, max=1.00000000)
    if(debug):
        test = residual(params, x, meas)
        print params, x, meas, test
    out = minimize(residual, params, args=(x,meas))
    out.leastsq()
    e_elem = out.params['e_elem'].value
    f_elem = out.params['f_elem'].value
    if(debug):
        report_errors(params)
        report_fit(out.params)
    return e_elem, f_elem
#=====================================================================

#=====================================================================
#  Log fitting 
#=====================================================================
def log_fit(x, a, b):
    return ( a * np.log2(x) + b )

def log_fit2(procs, var_elem):
    e_elem = var_elem[1]
    f_elem = var_elem[0]
    return ( f_elem * np.log2(procs) + e_elem )

def res_log(params, procs, measured):
    e_elem = params['e_elem'].value
    f_elem = params['f_elem'].value
    var_elem = [f_elem, e_elem]
    model = log_fit2(procs, var_elem)
    if (debug):
        print sum((measured-model)*(measured-model))
    return (measured - model) 

def Min_log(meas, x):
    params = Parameters()
    params.add('e_elem', value= 0.00000000)
    params.add('f_elem', value= 0.00000000)
    if(debug):
        test = res_log(params, x, meas)
        print params, x, meas, test
    out = minimize(res_log, params, args=(x,meas))
    out.leastsq()
    e_elem = out.params['e_elem'].value
    f_elem = out.params['f_elem'].value
    if(debug):
        report_errors(params)
        report_fit(out.params)
    return e_elem, f_elem

#=====================================================================

#=====================================================================
#  Minimum fitting 
#=====================================================================
def Min_fit(elem_list):
    min_list = min(elem_list)
    return min_list
#=====================================================================

#=====================================================================
#  Average fitting 
#=====================================================================
def Avg_fit(elem_list):
    avg = np.mean(elem_list)
    return avg
#=====================================================================

#=====================================================================
#  Print model factors independent variables
#=====================================================================
def print_model_factors_fitted(*args):
    i = 0
    Tcpu_cst = args[0]
    if (len(args) == 11):
        LB_0 = args[1]
        f_LB = args[2]
        i = 3
    elif (len(args) == 10):
        LB_cst = args[1]
        i = 2
    else:
        print "Some parameter is missing!!!! --- PRINT_MODEL"
      
    Ser_0 = args[0+i]
    f_ser = args[1+i]
    Trf_0 = args[2+i]
    f_trf = args[3+i]
    EffIPC_0 = args[4+i]
    f_effipc = args[5+i]
    EffInst_0 = args [6 + i]
    f_effinst = args[7+i]

    print "Calculating scalability model..."
    print "----Tcpu   = %.8f "%Tcpu_cst
    print "----Ser_0 = %.8f "%Ser_0," f_ser = %.8f"%f_ser
    print "----Trf_0  = %.8f "%Trf_0," f_trf = %.8f"%f_trf
    if (len(args) == 10):
        print "----LB      = %.8f "%LB_cst
    elif (len(args) == 11):
        print "----LB_0 = %.8f "%LB_0," f_LB = %.8f"%f_LB
    print "----EffIPC_0 = %.8f "%EffIPC_0," f_effipc = %.8f"%f_effipc
    print "----EffInst_0  = %.8f "%EffInst_0," f_effinst = %.8f"%f_effinst
    print "...............................................DONE!"
#=====================================================================

#==============================================================================================================
# NEW FUNCTION TO EVALUATE PARAMS.... P CUBIC OR LINEAR SERIALIZATION AMDAHL OR PIPELINE
#==============================================================================================================
def  projection_param(Proc_list, Ser_meas, Trf_meas, LB_meas, Eff_meas, 
                     EffIPC_meas, EffInst_meas,  Tcpu_meas, min_procs, 
                     p_ser, fit_ser, p_trf, fit_trf, p_lb, fit_lb, p_ipc, fit_IPC, p_inst, fit_Inst):
    
    print "_____________________________________________________________"
    print "============================================================="
    print "Projection based on input parameters \n"
    print "Serialization:"
    print "Processes: %s, Fitting Model: %s" %(p_ser, fit_ser)
    print "Transfer:"
    print "Processes: %s, Fitting Model: %s" %(p_trf, fit_trf)
    print "Load Balance:"
    print "Processes: %s, Fitting Model: %s" %(p_lb, fit_lb)
    print "Efficiency IPC:"
    print "Processes: %s, Fitting Model: %s" %(p_ipc, fit_IPC)
    print "Efficiency Inst:"
    print "Processes: %s, Fitting Model: %s" %(p_inst, fit_Inst)
    print "_____________________________________________________________"
    print "============================================================="    
    
    #d  = {}
    projected = []
    
    x  = np.array(Proc_list)

    Ser_0, f_ser = get_initial_values(p_ser, fit_ser, x, Ser_meas, 'Serialization')
    res_ser = [f_ser, Ser_0]
    
    Trf_0, f_trf = get_initial_values(p_trf, fit_trf, x, Trf_meas, 'Transfer')
    res_trf = [f_trf, Trf_0]

    if (fit_lb == 'avg') or (fit_lb == 'min'):
        LB_cst = get_initial_values_single(fit_lb, LB_meas, 'Load Balance')
    else:
        LB_0, f_LB = get_initial_values(p_lb, fit_lb, x, LB_meas, 'Load Balance')
        res_lb = [f_LB, LB_0]
    
    EffIPC_0, f_effipc = get_initial_values(p_ipc, fit_IPC, x, EffIPC_meas, 'IPC')
    res_effIPC = [f_effipc, EffIPC_0]
    
    EffInst_0, f_effinst = get_initial_values(p_inst, fit_Inst, x, EffInst_meas, 'Instructions')
    res_effInst = [f_effinst, EffInst_0]
    
#=========================================
# SUMMARY
#=========================================
    Tcpu_cst     = Avg_fit(Tcpu_meas)
   
    if (fit_lb == 'amdahl') or (fit_lb == 'log'):
        args = (Tcpu_cst, LB_0, f_LB, Ser_0, f_ser, Trf_0, f_trf, EffIPC_0, f_effipc, EffInst_0, f_effinst)
    else:
        args = (Tcpu_cst, LB_cst, Ser_0, f_ser, Trf_0, f_trf, EffIPC_0, f_effipc, EffInst_0, f_effinst)

    if(debug):
        print args
    print_model_factors_fitted(*args)

#=========================================
# EXTRAPOLACION
#=========================================
    i = 0
    procs = min_procs
     
    while (procs <= np.power(2, 20)):
        
        Ser = projection (p_ser, fit_ser, procs, res_ser, 'Serialization')
        Trf = projection (p_trf, fit_trf, procs, res_trf, 'Transfer')
        EffIPC = projection(p_ipc, fit_IPC, procs, res_effIPC, 'IPC')
        EffInst = projection(p_inst, fit_Inst, procs, res_effInst, 'Instructions')
        
        if (fit_lb == 'min') or (fit_lb == 'avg'):
            LB = LB_cst
        else:
            LB = projection (p_lb, fit_lb, procs, res_lb, 'Load Balance')
        
        #originally, IPC and Instructions were projected using lineal amdahl
        #   
        #P = procs    
        #EffIPC = f_amdahl(P, res_effIPC)
        #EffInst = f_amdahl(P, res_effInst)
         
        Eff = LB*Ser*Trf                     #Expected efficiency
        GlobalEff = Eff * EffIPC * EffInst
        Pcomm = 0.0       #Expected execution time based on ParEff
#         Pcomm = float(Tcpu/Eff)       #Expected execution time based on ParEff
        #TODO: Should we estimate Processing time with Global Efficiency
        #Pcomm = float(Tcpu/GlobalEff) 
        met = []
        met = [procs, Eff, LB, Ser, Trf, Pcomm, EffInst, EffIPC, GlobalEff]
        projected.append(met)
        i = i + 1
        procs *= 2

      
    return (projected)
#--------------------------------------------------------------------------------------------------------------
#==============================================================================================================
#					END OF THE FILE 
#==============================================================================================================
#--------------------------------------------------------------------------------------------------------------
